// This work is published under the GNU Public License (GPL)
// see file COPYING for details.
//
// The initial creation of this ACD application has been
// sponsored by IPTEL Inc. www.iptel.com.
//
// Author: Jan Willamowius <jan@willamowius.de>

package org.gnugk.acd;

import org.gnugk.*;
import java.util.Vector;
import java.util.Date;    // stay with Java 1.1 containers for now, so MS JView can run the app

/**
 * This is the main class for the ACD application.
 * It stores the agent data and implements the routing algorithms.
 *
 * @author Jan Willamowius <jan@willamowius.de>
 * @version 1.1
 */

public class GnuGkACD implements CallRouter {
    private static GnuGkACD instance;   // Singleton

    private Vector aliasList = new Vector();
	private GkClient gkclient;
    private ACDCommandListener commandListener;
    private ACDConfig config;
    private int distribution;   // call distribution algorithm
    private String gkHost;
    private int gkPort;
    private int listenPort;
    private String redirLocation;

    /**
     * Start the ACD.
     * @param args  unused
     */
	public static void main(String[] args) {
		if (args.length > 0) {
			System.out.println("usage: java org.gnugk.acd.GnuGkACD\n");
			System.exit(1);
		}		

		instance = new GnuGkACD();
	}

    /**
     * Singleton access.
     * @return singleton object
     */
    public static GnuGkACD getInstance() {
        return instance;
    }

    private GnuGkACD() {
        // read ACD config
        config = ACDConfig.getInstance();
        gkHost = config.getConfigKey("gnugk_host", "127.0.0.1");
        gkPort = config.getConfigKey("gnugk_port", 7000);
        listenPort = config.getConfigKey("listen_port", 7001);
        redirLocation = config.getConfigKey("redir_location", "http://www.google.com/");
        distribution = config.getDistributionAlgorithm();
        // connect to the gatekeeper
        gkclient = new GkClient();
		gkclient.init(this, gkHost, gkPort);
        String logfile = config.getConfigKey("logfile", "");
        int debugLevel = config.getConfigKey("debug", 0);
        Logger.setDebug(debugLevel);
        Logger.log("--- GnuGk ACD config ---");
        Logger.log("GnuGk host: " + gkHost);
        Logger.log("GnuGk status port: " + gkPort);
        Logger.log("ACD listen port: " + listenPort);
        Logger.log("ACD distribution algo: " + distribution);
        Logger.log("ACD redir location: " + redirLocation);
        Logger.log("ACD logfile: " + logfile);
        Logger.log("ACD debug level: " + debugLevel);
        Logger.log("--- starting ACD ---");
        if (logfile.length() > 0) {
            Logger.setLogfile(logfile);
        }
		new Thread(gkclient).start();
        // start listening for commands to the ACD
        commandListener = new ACDCommandListener();
        commandListener.init(listenPort, redirLocation);
        new Thread(commandListener).start();
    }

	public void addAlias(String _alias, String _epid) {
		if (aliasById(_alias, _epid) == null) {
			aliasList.addElement(new Agent(_epid, _alias));
		}
	}

    public void removeAllAliases(String _epid) {
		for(int i = 0; i < aliasList.size(); i++ ) {
			Alias alias = (Alias) aliasList.elementAt(i);
			if (_epid.equals(alias.getEpid())) {
				aliasList.removeElementAt(i);
			}
		}
	}

    /**
     * Set _all_ aliases on this endpoint with this CRV to available
     * @param _epid endpoint ID
     * @param _crv  CRV
     */
    public void setAliasAvailable(String _epid, String _crv) {
        Alias alias = aliasByCrv(_epid, _crv);  // find first alias
        while ((alias != null) && (_crv.length() > 0)) {
            alias.setState(Alias.AVAILABLE);
            ((Agent)alias).setLastCall();
            alias.setCrv("");
            alias = aliasByCrv(_epid, _crv);    // find next alias
        }
    }

    /**
     * Set this alias on this endpoint to TALKING and remember the CRV
     * @param _alias    H.323 alias
     * @param _crv      CRV
     */
    public void setAliasTalking(String _alias, String _crv) {
        Alias alias = aliasByName(_alias);
        if (alias != null) {
            alias.setState(Alias.TALKING);
            alias.setCrv(_crv);
        }
	}

    public void clearAllAliasStates() {
        aliasList.clear();
 	}

    private Alias aliasByName(String _alias) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Alias alias = (Alias) aliasList.elementAt(i);
            if (_alias.equals(alias.getAlias())) {
                return alias;
            }
        }
        return null;
    }

    private Alias aliasById(String _alias, String _epid) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Alias alias = (Alias) aliasList.elementAt(i);
            if (_alias.equals(alias.getAlias()) && _epid.equals(alias.getEpid())) {
                return alias;
            }
        }
        return null;
    }

    private Alias aliasByCrv(String _epid, String _crv) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Alias alias = (Alias) aliasList.elementAt(i);
            if (_epid.equals(alias.getEpid()) && _crv.equals(alias.getCrv()) ) {
                return alias;
            }
        }
        return null;
    }
	
    public Alias routeCall(String queue, String callerEndId,
                              String callRef, String callerAlias, String callerIp) {

        Agent agent = null;
        if (distribution == ACDConfig.FIRST_FIT) {
            agent = firstFitRouter(queue);
        } else if (distribution == ACDConfig.ROUND_ROBIN) {
            agent = roundRobinRouter(queue);
        } else if (distribution == ACDConfig.LONGEST_IDLE) {
            agent = longestIdleRouter(queue);
        }
        if (agent != null) {
            agent.setState(Agent.TALKING);
            agent.setLastCall();
        }
        if (agent != null) {
            Logger.log("Routing call for " + queue + " to " + agent);
        } else {
            Logger.log("Rejecting call for " + queue);
        }
        return agent;
    }

    // Algo: first-fit (unfair)
    private Agent firstFitRouter(String queue) {
        return firstFitRouter(0, queue);
    }

    // Algo: first-fit (unfair)
    private Agent firstFitRouter(int start, String queue) {
        if (start > aliasList.size()) {
            start = 0;
        }
        for(int i = start; i < aliasList.size(); i++ ) {
            Agent agent = (Agent) aliasList.elementAt(i);
            if (agent.isAvailable(queue)) {
                return agent;
            }
        }
        return null;
    }

    // Algo: round-robin (fair)
    private Agent lastPick = null;
    private Agent roundRobinRouter(String queue) {
        Agent agent = null;
        if (lastPick != null) {
            // find the agent we picked last
            for(int i = 0; i < aliasList.size(); i++ ) {
                if (lastPick.equals( aliasList.elementAt(i)) ) {
                    // find the next available agent (position to list end)
                    agent = firstFitRouter(i + 1, queue);
                    break;
                }
            }
        }
        if (agent == null) {
            // go around to the start and look again
            agent = firstFitRouter(0, queue);
        }
        lastPick = agent;
        return agent;
    }

    // Algo: longest-idle (fair)
    private Agent longestIdleRouter(String queue) {
        // pick the first available agent
        Agent longestIdle = firstFitRouter(0, queue);
        // see if we can find anyone that has been idle longer
        if (longestIdle != null) {  // no need to look if everybody is talking
            for(int i = 0; i < aliasList.size(); i++ ) {
                Agent agent = (Agent) aliasList.elementAt(i);
                if (agent.isAvailable(queue) &&
                   (agent.getLastCall() < longestIdle.getLastCall())) {
                    longestIdle = agent;
                }
            }
        }
        return longestIdle;
    }

    /**
     * agent is not available for ACD work.
     * @param agent
     */
    public void logout(String agent) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Agent ag = (Agent) aliasList.elementAt(i);
            if (agent.equals(H323Utils.extractAliasName(ag.getAlias()))) {
                ag.setBlocked(true);
                Logger.log("Logout agent " + agent);
            }
        }
        Logger.debug(dumpAgentStates());
    }

    /**
     * set agent to be available for ACD work again.
     * @param agent
     */
    public void login(String agent) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Agent ag = (Agent) aliasList.elementAt(i);
            if (agent.equals(H323Utils.extractAliasName(ag.getAlias()))) {
                ag.setBlocked(false);
                Logger.log("Login agent " + agent);
            }
        }
        Logger.debug(dumpAgentStates());
    }

    public String dumpAgentStates () {
        String result = "Agent list (" + aliasList.size() + " entries)\n";
        for(int i = 0; i < aliasList.size(); i++ ) {
            Agent ag = (Agent) aliasList.elementAt(i);
            if (ag.isAgent()) {
                // agent list contains all GK registration, only list configured agents
                result += ag.getAlias() + " (" + ag.getEpid() + "):\t" +
                        (ag.isBlocked() ? "NOT logged in" : "    logged in") +
                        "\tstate: " + (ag.getState()==0 ? "AVAILABLE" : "TALKING   " ) +
                        "\tlast call: " + new Date(ag.getLastCall()) +
                        "\n";
            }
        }
        return result;
    }

}
